@@ -6,11 +6,13 @@ use rustc_hash::FxHashMap;
 use test_utils::{
     extract_range_or_offset, Fixture, RangeOrOffset, CURSOR_MARKER, ESCAPED_CURSOR_MARKER,
 };
+use tt::Subtree;
 use vfs::{file_set::FileSet, VfsPath};
 
 use crate::{
     input::CrateName, Change, CrateDisplayName, CrateGraph, CrateId, Edition, Env, FileId,
-    FilePosition, FileRange, SourceDatabaseExt, SourceRoot, SourceRootId,
+    FilePosition, FileRange, ProcMacro, ProcMacroExpander, ProcMacroExpansionError,
+    SourceDatabaseExt, SourceRoot, SourceRootId,
 };
 
 pub const WORKSPACE: SourceRootId = SourceRootId(0);
@@ -81,7 +83,7 @@ pub struct ChangeFixture {
 
 impl ChangeFixture {
     pub fn parse(ra_fixture: &str) -> ChangeFixture {
-        let (mini_core, fixture) = Fixture::parse(ra_fixture);
+        let (mini_core, proc_macros, fixture) = Fixture::parse(ra_fixture);
         let mut change = Change::new();
 
         let mut files = Vec::new();
@@ -92,6 +94,7 @@ impl ChangeFixture {
         let mut default_cfg = CfgOptions::default();
 
         let mut file_set = FileSet::default();
+        let mut current_source_root_kind = SourceRootKind::Local;
         let source_root_prefix = "/".to_string();
         let mut file_id = FileId(0);
         let mut roots = Vec::new();
@@ -118,8 +121,13 @@ impl ChangeFixture {
                 assert!(meta.krate.is_some(), "can't specify deps without naming the crate")
             }
 
-            if meta.introduce_new_source_root {
-                roots.push(SourceRoot::new_local(mem::take(&mut file_set)));
+            if let Some(kind) = &meta.introduce_new_source_root {
+                let root = match current_source_root_kind {
+                    SourceRootKind::Local => SourceRoot::new_local(mem::take(&mut file_set)),
+                    SourceRootKind::Library => SourceRoot::new_library(mem::take(&mut file_set)),
+                };
+                roots.push(root);
+                current_source_root_kind = *kind;
             }
 
             if let Some(krate) = meta.krate {
@@ -197,7 +205,45 @@ impl ChangeFixture {
                 crate_graph.add_dep(krate, CrateName::new("core").unwrap(), core_crate).unwrap();
             }
         }
-        roots.push(SourceRoot::new_local(mem::take(&mut file_set)));
+
+        if !proc_macros.is_empty() {
+            let proc_lib_file = file_id;
+            file_id.0 += 1;
+
+            let (proc_macro, source) = test_proc_macros(&proc_macros);
+            let mut fs = FileSet::default();
+            fs.insert(
+                proc_lib_file,
+                VfsPath::new_virtual_path("/sysroot/proc_macros/lib.rs".to_string()),
+            );
+            roots.push(SourceRoot::new_library(fs));
+
+            change.change_file(proc_lib_file, Some(Arc::new(source)));
+
+            let all_crates = crate_graph.crates_in_topological_order();
+
+            let proc_macros_crate = crate_graph.add_crate_root(
+                proc_lib_file,
+                Edition::Edition2021,
+                Some(CrateDisplayName::from_canonical_name("proc_macros".to_string())),
+                CfgOptions::default(),
+                CfgOptions::default(),
+                Env::default(),
+                proc_macro,
+            );
+
+            for krate in all_crates {
+                crate_graph
+                    .add_dep(krate, CrateName::new("proc_macros").unwrap(), proc_macros_crate)
+                    .unwrap();
+            }
+        }
+
+        let root = match current_source_root_kind {
+            SourceRootKind::Local => SourceRoot::new_local(mem::take(&mut file_set)),
+            SourceRootKind::Library => SourceRoot::new_library(mem::take(&mut file_set)),
+        };
+        roots.push(root);
         change.set_roots(roots);
         change.set_crate_graph(crate_graph);
 
@@ -205,6 +251,41 @@ impl ChangeFixture {
     }
 }
 
+fn test_proc_macros(proc_macros: &[String]) -> (Vec<ProcMacro>, String) {
+    // The source here is only required so that paths to the macros exist and are resolvable.
+    let source = r#"
+#[proc_macro_attribute]
+pub fn identity(_attr: TokenStream, item: TokenStream) -> TokenStream {
+    item
+}
+#[proc_macro_attribute]
+pub fn input_replace(attr: TokenStream, _item: TokenStream) -> TokenStream {
+    attr
+}
+"#;
+    let proc_macros = std::array::IntoIter::new([
+        ProcMacro {
+            name: "identity".into(),
+            kind: crate::ProcMacroKind::Attr,
+            expander: Arc::new(IdentityProcMacroExpander),
+        },
+        ProcMacro {
+            name: "input_replace".into(),
+            kind: crate::ProcMacroKind::Attr,
+            expander: Arc::new(AttributeInputReplaceProcMacroExpander),
+        },
+    ])
+    .filter(|pm| proc_macros.iter().any(|name| name == pm.name))
+    .collect();
+    (proc_macros, source.into())
+}
+
+#[derive(Debug, Clone, Copy)]
+enum SourceRootKind {
+    Local,
+    Library,
+}
+
 #[derive(Debug)]
 struct FileMeta {
     path: String,
@@ -213,7 +294,7 @@ struct FileMeta {
     cfg: CfgOptions,
     edition: Edition,
     env: Env,
-    introduce_new_source_root: bool,
+    introduce_new_source_root: Option<SourceRootKind>,
 }
 
 impl From<Fixture> for FileMeta {
@@ -229,7 +310,41 @@ impl From<Fixture> for FileMeta {
             cfg,
             edition: f.edition.as_ref().map_or(Edition::CURRENT, |v| Edition::from_str(v).unwrap()),
             env: f.env.into_iter().collect(),
-            introduce_new_source_root: f.introduce_new_source_root,
+            introduce_new_source_root: f.introduce_new_source_root.map(|kind| match &*kind {
+                "local" => SourceRootKind::Local,
+                "library" => SourceRootKind::Library,
+                invalid => panic!("invalid source root kind '{}'", invalid),
+            }),
         }
     }
 }
+
+// Identity mapping
+#[derive(Debug)]
+struct IdentityProcMacroExpander;
+impl ProcMacroExpander for IdentityProcMacroExpander {
+    fn expand(
+        &self,
+        subtree: &Subtree,
+        _: Option<&Subtree>,
+        _: &Env,
+    ) -> Result<Subtree, ProcMacroExpansionError> {
+        Ok(subtree.clone())
+    }
+}
+
+// Pastes the attribute input as its output
+#[derive(Debug)]
+struct AttributeInputReplaceProcMacroExpander;
+impl ProcMacroExpander for AttributeInputReplaceProcMacroExpander {
+    fn expand(
+        &self,
+        _: &Subtree,
+        attrs: Option<&Subtree>,
+        _: &Env,
+    ) -> Result<Subtree, ProcMacroExpansionError> {
+        attrs
+            .cloned()
+            .ok_or_else(|| ProcMacroExpansionError::Panic("Expected attribute input".into()))
+    }
+}
